/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2003 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: jorm-team@objectweb.org
 *
 */

package org.objectweb.jorm.compiler.api;

import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.util.io.api.PathExplorer;
import org.objectweb.jorm.util.api.Loggable;

import java.io.InputStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;

/**
 * A JormCompilerParameter object is an object which manages informations
 * read by the command line parser. These informations are needed to
 * the compiler system to parse, generate, and compile.
 * @author	X. Spengler
 */
public interface JormCompilerParameter extends Cloneable, Loggable {
	/**
	 * The pattern that represents the JORM class name to be replaced into
	 * the string defining the inherited class of a either the PBinding or the
	 * PClassMapping that is generated.
	 */
	public static final String INHERITANCE_CLASSNAME_PATTERN = "%c";
	public static final String INHERITANCE_PACKAGE_PATTERN = "%p";

	/**
	 * Prints informations managed by the current object.
	 */
	public void print();

	/**
	 * Computes the actual class name which the generated PClassMapping class
	 * must extend.
	 * @param cn	The class name of the persistent class for which the
	 * 				code is generated.
	 * @return	The computed class name.
	 */
	public String computePClassMappingInheritance(String cn);

	/**
	 * Computes the actual class name which the generated PBinding class
	 * must extend.
	 * @param cn	The class name of the persistent class for which the
	 * 				code is generated.
	 * @return	The computed class name.
	 */
	public String computePBindingInheritance(String cn);

	/**
	 * Load a jorm configuration file.
	 * @param file  the name of the jorm configuration file
	 */
	public void loadConfFile(String file, Iterator knownmappers) throws PException;

	/**
	 * Load a jorm configuration file.
	 * @param in  the input stream of the jorm configuration file
	 */
	public void loadConfFile(InputStream in, Iterator knownmappers) throws PException;

	public PathExplorer getClasspath();

	public String getOutput();

	public boolean isVerbose();

	public Collection getInputFiles();

	public boolean isKeepSrc();

	public String getJavacName();

	public String getProjectName();

	public boolean isJavac();

	public boolean isParseOnly();

	public String getBindingInheritance();

	public String getClassMappingInheritance();

	public boolean isBindingAbstract();

	public ArrayList getDtdLocations();

	public boolean isGeneratedPDFiles();

	public boolean isGeneratedWithMapperPackage();

	public void setClasspath(PathExplorer classpath);

	public void setOutput(String output);

	public void setVerbose(boolean verbose);

	public void setKeepSrc(boolean keepSrc);

	public void setProjectName(String projectName);

	public void setJavac(boolean javac);

	public void setBindingInheritance(String bindingInheritance);

	public void setClassMappingInheritance(String classMappingInheritance);

	public void setBindingAbstract(boolean bindingAbstract);

	public void setDtdLocations(ArrayList dtdLocations);

	public void setGeneratedPDFiles(boolean generatedPDFiles);

	public void setInputFiles(Collection inputFiles);

	public void setGeneratedWithMapperPackage(boolean generatedWithMapperPackage);

	public void setLogConfFile(String logConfFile);

	public Object clone();
}
