/**
 * JORM: an implementation of a generic mapping system for persistent Java
 *       objects. Two mapping are supported: to RDBMS and to binary files.
 * Copyright (C) 2001-2003 France Telecom R&D - INRIA
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: jorm-team@objectweb.org
 *
 */

package org.objectweb.jorm.mapper.rdb.generator;

import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.compiler.api.JormCompilerParameter;
import org.objectweb.jorm.generator.api.MOP;
import org.objectweb.jorm.generator.api.MOPFactory;
import org.objectweb.jorm.metainfo.api.*;
import org.objectweb.jorm.metainfo.api.Package;
import org.objectweb.jorm.metainfo.api.Class;
import org.objectweb.jorm.mapper.rdb.metainfo.*;
import org.objectweb.jorm.util.api.Loggable;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.LoggerFactory;

import java.util.*;

/**
 *
 */
public class RdbMOPFactory implements MOPFactory, Loggable {


    private HashMap mapping2geninfos = new HashMap();
	private Logger logger;
	private LoggerFactory loggerFactory;
	private boolean debug = false;

	// IMPLEMENTATION OF METHODS FROM THE MOPFactory INTERFACE //
	//---------------------------------------------------------//

	public Logger getLogger() {
		return logger;
	}

	public LoggerFactory getLoggerFactory() {
		return loggerFactory;
	}

	public void setLogger(Logger logger) {
		this.logger = logger;
	}

	public void setLoggerFactory(LoggerFactory loggerfactory) {
		this.loggerFactory = loggerfactory;
		if (logger == null && this.loggerFactory != null) {
			logger = this.loggerFactory.getLogger(getClass().getName());
		}
	}

	// IMPLEMENTATION OF METHODS FROM THE MOPFactory INTERFACE //
	//---------------------------------------------------------//

	/**
	 * This method must return a new instance of a MappingMOP
	 */
	public MOP createMappingMOP(JormCompilerParameter cp, String mappername) throws PException {
		RdbMappingMOP res = new RdbMappingMOP();
		return res;
	}

	/**
	 * This method must return a new instance of a BindingMOP
	 */
	public MOP createBindingMOP(JormCompilerParameter cp, String mappername) throws PException {
		RdbBindingMOP res = new RdbBindingMOP();
		return res;
	}

	/**
	 * Creates a generation context that can be shared between binding and
	 * mapping generation for a particular class.
     *
     * If the current class has a parent class and if a vertical mapping
     * has been defined: the main table of the current class then references
     *
     *
     * cm.getRdbTable() returns the main table of one of the parent classes (P class).
     * rcm.getRdbAllExternalTables() : returns all the external tables defines for
     * the current class and for all the parent classes up to the P class.
     *
	 */
	public Object createClassGenerationContext(Mapping mo) throws PException {
		debug = logger != null && logger.isLoggable(BasicLevel.DEBUG);
        if (mapping2geninfos.containsKey(mo)) {
            return mapping2geninfos.get(mo);
        }
		RdbGenInfos rgi = new RdbGenInfos();
        mapping2geninfos.put(mo, rgi);
		RdbClassMapping currentcm = (RdbClassMapping) mo.getClassMapping();
		if (debug) {
			logger.log(BasicLevel.DEBUG, "mo.getClassMapping() " + mo.getClassMapping());
			logger.log(BasicLevel.DEBUG, "Creating the generation MI for the class = "
			    + ((Class)currentcm.getLinkedMO()).getFQName());
		}

		// define the main table
        ArrayList mainTablesList = new ArrayList();
        currentcm.getMainRdbTable(mainTablesList);
        // the only case currently handled: one main table is retrieved
        RdbTable mainrt = (RdbTable) mainTablesList.get(0);
		if (debug) {
			logger.log(BasicLevel.DEBUG, "Main table: " + mainrt.getName());
		}
        RdbClassMapping maincm = (RdbClassMapping) mainrt.getParent();
		RdbGenTable mainTable = new RdbGenTable(mainrt);
	    //logger.log(BasicLevel.DEBUG, "main table =" + mainTable.getTableName());
		mainTable.genInfos = rgi;
		rgi.referenceTable = mainTable;

        // Get first the inherited columns
        ArrayList inheritedTableColumns = new ArrayList();
        ArrayList inheritedReadableColumns = new ArrayList();
        Map inheritedGenRefs = new HashMap();
        Map inheritedExtTables = new HashMap();
        getInheritedData(maincm, inheritedTableColumns, inheritedReadableColumns, inheritedGenRefs, inheritedExtTables);
        if (!inheritedTableColumns.isEmpty()) {
            //mainTable.isMainTable = false;
            mainTable.filter = maincm.getRdbFilter();
            mainTable.inheritedColumns.addAll(inheritedTableColumns);
		    //rgi.readableColumns.addAll(mainTable.inheritedColumns);
            rgi.readableColumns.addAll(inheritedReadableColumns);
            rgi.genRefs.putAll(inheritedGenRefs);
            rgi.genTables.putAll(inheritedExtTables);
	        if (debug) {
		        logger.log(BasicLevel.DEBUG, "Inherited columns:");
		        printColumns(inheritedTableColumns);
	        }
        } // Filtered mapping and concrete parent classes

        // Get then the local data
        rgi.genTables.put(mainTable.tableName, mainTable);
		rgi.readableColumns.addAll(mainTable.tableColumns);

        // Get the external tables
        ArrayList extTablesList = new ArrayList();
		if (currentcm instanceof org.objectweb.jorm.mapper.rdb.metainfo.RdbClassMultiMapping) {
			RdbClassMultiMapping rcm = (RdbClassMultiMapping) currentcm;
            rcm.getAllRdbExternalTables(extTablesList);
			for (Iterator it = extTablesList.iterator(); it.hasNext();) {
		        RdbExternalTable moTable = (RdbExternalTable) it.next();
				RdbGenTable curTable = new RdbGenTable(moTable, mainTable);
                if (debug) {
		            logger.log(BasicLevel.DEBUG,
		                "ext table -- name = " + curTable.getTableName());
                }
				curTable.genInfos = rgi;
				rgi.genTables.put(curTable.tableName, curTable);
				for (int i = 0; i < curTable.tableColumns.size(); i++) {
			        RdbGenColumn rgc = (RdbGenColumn) curTable.tableColumns.get(i);
					if (rgc.joinCol == null) {
				        rgi.readableColumns.add(rgc);
					}
				 }
            }
		}

		// define GenRef of the class identifier
		//System.out.println("get the namedef");
		NameDef nd = (NameDef) currentcm.getIdentifierMapping().getLinkedMO();
        Class parentclass = (Class) nd.getParent();
        Class currentClass = (Class) currentcm.getLinkedMO();
     /*
        for (Iterator iter = currentClass.iterateAllField(); iter.hasNext();) {
            TypedElement typedElem = (TypedElement) iter.next();
            if (typedElem instanceof Reference) {
            System.out.println("current class: " + currentClass.getName() + "current reference field " +   typedElem.getName());
            System.out.println("currentcm.getReferenceMapping(typedElem.getName()) " +   currentcm.getReferenceMapping(typedElem.getName()));
            NameDef namedef = (NameDef) currentcm.getReferenceMapping(typedElem.getName()).getLinkedMO();
            System.out.println("current reference field " +   typedElem.getName() + " namedef " + namedef);
            }
        }
        */

        int ndId = 0;
        if (!currentClass.getFQName().equals(parentclass.getFQName()) &&
                // filtered case                  || ! filtered case && parent class is abstract
                (!inheritedTableColumns.isEmpty() || (inheritedTableColumns.isEmpty() && !parentclass.isAbstract()))){
            // The identifier is mapped at a upper level in the inheritance tree.
		    if (debug) {
		        logger.log(BasicLevel.DEBUG, "Namedef inherited from the class " + parentclass.getFQName());
		    }
            RdbClassMapping parentcm =
                            (RdbClassMapping) parentclass.getClassMapping(currentcm.getProjectName(),currentcm.getMapperName());
            rgi.genId = defineGenRef(nd, null, ndId++, currentcm, parentcm, rgi);

        } else {
            // The identifier is mapped locally.
            rgi.genId = defineGenRef(nd, null, ndId++, currentcm, null, rgi);
         }

		// define GenRef for each reference field of the current class
		for (Iterator it = currentcm.getReferenceMappings().iterator(); it.hasNext();) {
	        nd = (NameDef) ((ReferenceMapping) it.next()).getLinkedMO();
			String fn = ((Reference) nd.getParent()).getName();
			rgi.genRefs.put(fn, defineGenRef(nd, fn, ndId++, currentcm, null, rgi));
		 }

        //System.out.println("rgi.genTables.values().iterator()");
		for (Iterator it = rgi.genTables.values().iterator(); it.hasNext();) {
			((RdbGenTable) it.next()).setGenerateUpdate(rgi);
		}

		return rgi;
	}

	// PRIVATE METHODS

	/**
	 * Defines the RdbGenDef associated either with the identifier of this class
	 * mapping, or with one of its references.
	 * @param nd    The NameDef associated with the identifier or a reference.
	 * @param cnid  The composite name identifier allocated for code generation
	 *              of references.
	 * @return  The created RdbGenRef describing the identifier or a reference.
	 */
	private RdbGenRef defineGenRef(NameDef nd, String fn, int cnid,
								   RdbClassMapping cm,
	                               RdbClassMapping parentcm,
	                               RdbGenInfos rgi)
			throws PException {

		RdbGenRef rgr = new RdbGenRef();
		rgr.nd = nd;
		rgr.fieldName = fn;
		if (nd.isFieldName()) {
            RdbPrimitiveElementMapping pem = (RdbPrimitiveElementMapping)
                cm.getPrimitiveElementMapping(nd.getFieldName());
			if (pem == null && parentcm != null) {
                pem = (RdbPrimitiveElementMapping) parentcm
                    .getPrimitiveElementMapping(nd.getFieldName());
            }
			if (pem == null) {
				throw new PException(
				    "Unable to find the mapping of the field "
				    + nd.getFieldName());
			}
			RdbTable table = (RdbTable) pem.getParent();
			if (debug) {
				logger.log(BasicLevel.DEBUG, "nd(" + fn
				    + ") FieldName, field=" + nd.getFieldName()
				    + " / column name=" + pem.getName()
				    + " in the table: " + table.getName());
			}
			RdbGenTable rgt = (RdbGenTable) rgi.genTables.get(table.getName());
			if (rgt == null) {
				throw new NullPointerException("No table " + table.getName() + " found among: " + rgi.genTables);
			}
			rgr.refColumn = rgt.getColumn(pem.getName());
			if (rgt.joins != null) {
				getJoin(rgt, pem, nd.getFieldName(), rgr);
			}
		} else { // this is a NameRef
			if (debug) {
				logger.log(BasicLevel.DEBUG, "nd(" + fn + ") CommpositeName, cnId=" + cnid);
			}
			rgr.cnId = cnid;
			rgr.cnName = nd.getNameRef().getCompositeName().getName();
			rgr.cnPackage = ((Package) nd.getNameRef().getCompositeName().getParent()).getName();
			rgr.cnFieldColumns = new TreeMap();
			Iterator itp = nd.getNameRef().getProjection().entrySet().iterator();
            RdbPrimitiveElementMapping pem = null;
			while (itp.hasNext()) {
				Map.Entry me = (Map.Entry) itp.next();
				String cnfield = (String) me.getKey();
				String clafn = (String) me.getValue();
				if (parentcm != null) {
                    pem = (RdbPrimitiveElementMapping) parentcm.getPrimitiveElementMapping(clafn);
                } else {
                    pem = (RdbPrimitiveElementMapping) cm.getPrimitiveElementMapping(clafn);
                }
				if (pem == null) {
					throw new PException(
							"Unable to find the mapping of the field "
							+ me.getValue());
				}
				RdbTable table = (RdbTable) pem.getParent();
				if (debug) {
					logger.log(BasicLevel.DEBUG, "nd(" + fn
					    + ") composite field name=" + cnfield
					    + " /  class field name=" + clafn
					    + " / column name=" + pem.getName()
					    + " in the table: " + table.getName());
				}
				RdbGenTable rgt = (RdbGenTable) rgi.genTables.get(table.getName());
				RdbGenColumn c = rgt.getColumn(pem.getName());
				rgr.cnFieldColumns.put(cnfield, c);
				if (rgr.join == null && rgt.joins != null) {
					getJoin(rgt, pem, clafn, rgr);
				}
			}
		}
		return rgr;
	}

	private void getJoin(RdbGenTable rgt,
						 RdbPrimitiveElementMapping pem,
						 String fn,
						 RdbGenRef rgr) {
		if (debug) {
			logger.log(BasicLevel.DEBUG, "getJoin: fn: " + rgr.fieldName
			    + " / col: " + pem.getName()
			    + " / rgt: " + rgt.tableName);
		}
		if (rgt.joins.size() == 1) {
			rgr.join = (RdbGenJoin) rgt.joins.get(0);
			if (debug) {
				logger.log(BasicLevel.DEBUG, "return: fn: " + rgr.fieldName
				    + " / col: " + pem.getName()
				    + " / join: " + rgr.join.joinIdx);
			}
		}
		Iterator it = pem.getPrimitiveElementByRdbJoin().entrySet().iterator();
		while (it.hasNext()) {
			Map.Entry me2 = (Map.Entry) it.next();
			PrimitiveElement pe = (PrimitiveElement) me2.getValue();
			if (pe.getName().equals(fn)) {
				rgr.join = (RdbGenJoin)
						rgt.jn2join.get(((RdbJoin) me2.getKey()).getName());
				if (debug) {
					logger.log(BasicLevel.DEBUG, "return fn: " + rgr.fieldName
					    + " / col: " + pem.getName()
					    + " / joinName: " + ((RdbJoin) me2.getKey()).getName()
					    + " / join: " + rgr.join.joinIdx);
				}
			}
		}
	}

    private void getInheritedData(RdbClassMapping cm, ArrayList inheritedTableColumns, ArrayList inheritedReadableColumns,
                                  Map inheritedGenRefs, Map inheritedExtTables) throws PException {
		for (Iterator pcmit = cm.getParentClassMappings().iterator(); pcmit.hasNext();) {
	        ParentClassMapping pcm = (ParentClassMapping) pcmit.next();
            if (pcm.getRuleName().equals("filtered")) {
                Class parentClass = pcm.getMOClass();
                if (debug) {
					logger.log(BasicLevel.DEBUG,
					    "rule-name = filtered =<" + parentClass.getFQName() + ">"
					    + " abstract =<" + parentClass.isAbstract() + ">");
                }
               // if (! parentClass.isAbstract()) {
                    RdbClassMapping parentcm =
                            (RdbClassMapping) parentClass.getClassMapping(cm.getProjectName(),cm.getMapperName());
                    RdbGenInfos rgi = (RdbGenInfos) createClassGenerationContext((Mapping) parentcm.getParent());
                    RdbGenTable temp = rgi.getReferenceTable();
                    inheritedTableColumns.addAll(0, temp.tableColumns);
                    inheritedReadableColumns.addAll(0, rgi.readableColumns);
                    inheritedGenRefs.putAll(rgi.genRefs);
                    inheritedExtTables.putAll(rgi.genTables);
                    getInheritedData(parentcm, inheritedTableColumns, inheritedReadableColumns,
                                        inheritedGenRefs, inheritedExtTables);
               // }
                /* else {
                    // Look up for the first concrete class from the parentClass
                    Class aParentClass = lookupConcreteClass(parentClass, cm);
                    if (aParentClass != null) {
                        RdbClassMapping parentcm =
                                (RdbClassMapping) parentClass.getClassMapping(cm.getProjectName(),cm.getMapperName());
                        RdbGenInfos rgi = (RdbGenInfos) createClassGenerationContext((Mapping) parentcm.getParent());
                        RdbGenTable temp = rgi.getReferenceTable();
                        //TODO : cyclic graph ?
                        inheritedTableColumns.addAll(0, temp.tableColumns);
                        inheritedReadableColumns.addAll(0, rgi.readableColumns);
                        inheritedGenRefs.putAll(rgi.genRefs);
                        inheritedExtTables.putAll(rgi.genTables);
                        getInheritedData(parentcm, inheritedTableColumns, inheritedReadableColumns,
                                            inheritedGenRefs, inheritedExtTables);
                    }
                }     */
            }
        }
    }

    // Helper method
    // TODO : only filtered mapping is allowed up to the concrete class
   private Class lookupConcreteClass(Class fromClass, RdbClassMapping cm) throws PException {
       Collection parentClasses = fromClass.getExtensions();
        // the only case currently handled: there is one parent class
       if (!parentClasses.isEmpty()) {
           Class parentClass = (Class) parentClasses.iterator().next();
            RdbClassMapping parentcm =
                            (RdbClassMapping) parentClass.getClassMapping(cm.getProjectName(),cm.getMapperName());
           if (parentcm == null) {
               if (!parentClass.isAbstract()) {
                    return parentClass;
               } else {
                lookupConcreteClass(parentClass, cm);
               }
           } // filtered mapping
       }
       return null;
   }

     private void printColumns(List columns) {
        RdbGenColumn rgc = null;
        for (int i = 0; i < columns.size(); i++) {
            rgc = (RdbGenColumn) columns.get(i);
            logger.log(BasicLevel.DEBUG, "\tcolumn : " + rgc.columnName
                + " isHiddenFilterField: " + rgc.isHiddenFilterField());
        }
    }
}
